package com.persagy.proxy.migration.handler;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.util.StrUtil;
import com.fasterxml.jackson.databind.node.JsonNodeFactory;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.persagy.dmp.basic.model.QueryCriteria;
import com.persagy.dmp.basic.utils.JsonNodeUtils;
import com.persagy.dmp.common.constant.ResponseCode;
import com.persagy.dmp.common.model.entity.BaseEntity;
import com.persagy.dmp.define.client.DigitalObjectInfoFacade;
import com.persagy.dmp.define.entity.ObjectInfoDefine;
import com.persagy.proxy.common.entity.InstanceUrlParam;
import com.persagy.proxy.migration.constant.MigrationConstant;
import com.persagy.proxy.migration.constant.MigrationType;
import com.persagy.proxy.migration.model.DataMigrationExcel;
import com.persagy.proxy.migration.model.DataMigrationResponse;
import com.persagy.proxy.migration.model.MigrationInfo;
import com.persagy.proxy.migration.service.Impl.MigrationAbstractServiceImpl;
import lombok.extern.slf4j.Slf4j;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * @author lvxianyun
 * @title: DefineInfoMigration
 * @projectName adm-middleware
 * @description:  迁移adm交付的数据到现场，覆盖现场数据，提供增加、删除、修改 记录放入excel; DefineInfo 数据迁移 adm -> target project
 * @date 2021/12/8 17:28
 */
@Slf4j
public class DefineInfoMigration extends MigrationAbstractServiceImpl<ObjectInfoDefine> implements IDataMigration, IDataMigrationProcess<ObjectInfoDefine> {
    
    private static DefineInfoMigration instance;

    private DefineInfoMigration() {}

    // 静态方法返回该实例
    public static DefineInfoMigration getInstance() {
        if(instance == null) {
            synchronized (DefineInfoMigration.class) {
                if (instance == null) {
                    instance = new DefineInfoMigration();
                }
            }
        }
        return instance;
    }


    /**
     * 数据迁移
     *
     * @param context
     * @param migrationInfo
     * @return
     */
    @Override
    public List<DataMigrationExcel> migrate(InstanceUrlParam context, MigrationInfo migrationInfo) {
        log.info("######################### DefineInfo 同步开始 #########################");
        long start = System.currentTimeMillis();
        List<DataMigrationExcel> dataMigrationExcels = new ArrayList<>();

        QueryCriteria queryCriteria = new QueryCriteria();
        ObjectNode criteria = JsonNodeFactory.instance.objectNode();
        criteria.put("valid", 1);

        ObjectNode groupCode =  JsonNodeFactory.instance.objectNode();
        groupCode.put("$ne", "0");

        ObjectNode projectId =  JsonNodeFactory.instance.objectNode();
        projectId.put("$ne", "0");

        ObjectNode andOr = JsonNodeFactory.instance.objectNode();
        andOr.putPOJO("groupCode", projectId);
        andOr.putPOJO("projectId", groupCode);

        criteria.putPOJO("$andOr", andOr);

        queryCriteria.setCriteria(criteria);
        queryCriteria.setSize(-1L);
        List<ObjectInfoDefine> admDefineList = getAdmData(context, queryCriteria);
        String queryUrl = requestUrl(context, migrationInfo, MigrationType.QUERY.getCode());
        List<ObjectInfoDefine> projectDefineList = getProjectData(queryCriteria, ObjectInfoDefine.class, queryUrl);

        if(CollUtil.isEmpty(admDefineList)){
            log.error(" ######################### DefineInfo adm已交付需要同步的数据:"+ ResponseCode.C0320.getDesc() +" 同步结束 #########################");
            return dataMigrationExcels;
        }
        String insertUrl = requestUrl(context, migrationInfo, MigrationType.INSERT.getCode());
        if(CollUtil.isEmpty(projectDefineList)){
            log.info("######################### DefineInfo 批量插入数据 #########################");
            DataMigrationResponse response = insertBatch(admDefineList, ObjectInfoDefine.class, insertUrl);
            dataMigrationExcels.addAll(processData(response, MigrationType.INSERT.getDesc()));
            return dataMigrationExcels;
        }
        List<ObjectNode> objectNodeListPro = JsonNodeUtils.toListNode(projectDefineList, null, null);
        Map<String,Object> projectDefineMap = toEntityMap(objectNodeListPro, ObjectInfoDefine.class);
        List<ObjectNode> objectNodeListAdm = JsonNodeUtils.toListNode(admDefineList, null, null);
        Map<String,Object> admDefineMap = toEntityMap(objectNodeListAdm, ObjectInfoDefine.class);


        //差集 新增
        log.info("######################### DefineInfo 批量插入数据 #########################");
        Map<String,Object> subtractFromDel = doSubtractAdd(admDefineMap, projectDefineMap);
        if(!CollUtil.isEmpty(subtractFromDel)){
            List<ObjectInfoDefine> insertData = toList(subtractFromDel, admDefineList);
            dataMigrationExcels.addAll(processData(insertBatch(insertData, ObjectInfoDefine.class, insertUrl), MigrationType.INSERT.getDesc()));
        }

        //差集 删除
        log.info("######################### DefineInfo 批量处理删除数据 #########################");
        Map<String,Object> subtractFromDelete = doSubtractDel(admDefineMap, projectDefineMap);
        if(!CollUtil.isEmpty(subtractFromDelete)){
            List<String> deleteIds = subtractFromDelete.entrySet().stream().map(en -> en.getKey()).collect(Collectors.toList());
            String delUrl = requestUrl(context, migrationInfo, MigrationType.DELETE.getCode());
            DataMigrationResponse response = deleteBatch(deleteIds, ObjectInfoDefine.class, delUrl);
            dataMigrationExcels.addAll(processData(response, MigrationType.DELETE.getDesc()));
        }

        //交集更新
        log.info("######################### DefineInfo 批量处理更新数据 #########################");
        Map<String,Object> intersectionUpdateFrom = doIntersectionGetFrom(admDefineMap, projectDefineMap);
        Map<String,Object> intersectionUpdateTo = doIntersectionGetTo(admDefineMap, projectDefineMap);

        if(!CollUtil.isEmpty(intersectionUpdateFrom) && !CollUtil.isEmpty(intersectionUpdateTo)){
            Map<String, Object> compareData = compareData(intersectionUpdateFrom, intersectionUpdateTo);
            if(CollUtil.isNotEmpty(compareData) && compareData.size() > 0){
                List<ObjectInfoDefine> updateData = toList(compareData, admDefineList);
                String updateUrl = requestUrl(context, migrationInfo, MigrationType.UPDATE.getCode());
                dataMigrationExcels.addAll(processData(updateBatch(updateData, ObjectInfoDefine.class, updateUrl), MigrationType.UPDATE.getCode()));
            }
        }
        long end = System.currentTimeMillis();
        log.info("######################### DefineInfo 数据迁移已结束 时间："+(end-start)+" #########################");

        return dataMigrationExcels;
    }

    /**
     * 获取adm交付的数据，默认获取有效数据
     *
     * @param context
     * @return
     */
    @Override
    public List<ObjectInfoDefine> getAdmData(InstanceUrlParam context, QueryCriteria queryCriteria) {
        log.info("######################### DefineInfo 获取adm交付数据 #########################");
        List<ObjectInfoDefine> infos = DigitalObjectInfoFacade.query(context.getGroupCode(),context.getProjectId(),context.getUserId(),context.getUserId(),queryCriteria);
        return infos;
    }

    /**
     * 根据map转为需要处理的对象集合
     * 使用 MigrationConstant.spliter 作为拼接符
     * @param fromMap
     * @param fromList
     * @return
     */

    public List<ObjectInfoDefine> toList(Map<String, Object> fromMap, List<ObjectInfoDefine> fromList) {
        if(CollUtil.isEmpty(fromMap) || CollUtil.isEmpty(fromList)){
            return null;
        }
        List<ObjectInfoDefine> results = new ArrayList<>(fromMap.size());
        Map<String, ObjectInfoDefine> objectInfoDefineMap = CollUtil.fieldValueMap(fromList, BaseEntity.PROP_ID);
        for(String key : fromMap.keySet()){
            results.add(objectInfoDefineMap.get(key));
        }
        return results;
    }

    /**
     * 拼接访问链接
     *
     * @param context
     * @param migrationInfo
     * @param type
     * @return
     */
    @Override
    public String requestUrl(InstanceUrlParam context, MigrationInfo migrationInfo, String type) {
        if(StrUtil.isBlank(migrationInfo.getTargetUrl())){
            return null;
        }
        StringBuilder url = new StringBuilder(migrationInfo.getTargetUrl());
        if(type.equals(MigrationType.QUERY.getCode())) {
            url.append(MigrationConstant.DEF_FUNCID);
        }else if(type.equals(MigrationType.INSERT.getCode())){
            url.append(MigrationConstant.DEF_FUNCID_INSERT);
        }else if(type.equals(MigrationType.UPDATE.getCode())){
            url.append(MigrationConstant.DEF_FUNCID_UPDATE);
        }else{
            url.append(MigrationConstant.DEF_FUNCID_DELETE);
        }
        return commonUrl(context, url.toString(), migrationInfo);
    }

    /**
     * 将目标集合处理成excel格式的集合  使用与insert/delete
     *
     * @param dataMigrationResponse
     * @return
     */
    @Override
    public List<DataMigrationExcel> processData(DataMigrationResponse dataMigrationResponse, String type) {
        if(CollUtil.isEmpty(dataMigrationResponse.getData()) && CollUtil.isEmpty(dataMigrationResponse.getErrorMsg())){
            return Collections.emptyList();
        }
        List<DataMigrationExcel> dataMigrationExcels = new ArrayList<>();
        if(CollUtil.isNotEmpty(dataMigrationResponse.getData())){
            for(Object obj : dataMigrationResponse.getData()){
                if(!type.equals(MigrationType.DELETE.getCode())){
                    ObjectInfoDefine define = (ObjectInfoDefine) obj;
                    DataMigrationExcel excel = new DataMigrationExcel();
                    excel.setId(define.getId());
                    excel.setCode(define.getCode());
                    excel.setName(define.getName());
                    excel.setType(type);
                    dataMigrationExcels.add(excel);
                }else{
                    DataMigrationExcel excel = new DataMigrationExcel();
                    excel.setId(obj.toString());
                    excel.setType(type);
                    dataMigrationExcels.add(excel);
                }
            }
        }
        if(CollUtil.isNotEmpty(dataMigrationResponse.getErrorMsg())){
            dataMigrationExcels.addAll(addErrorInfo(dataMigrationResponse.getErrorMsg()));
        }
        return dataMigrationExcels;
    }
}
